@@ -16,15 +16,14 @@ deriveArgDict n = do
   (typeHead, constrs) <- getDeclInfo n
   c <- newName "c"
   ts <- gadtIndices c constrs
-  let xs = flip map ts $ \case
-        Left t -> AppT (AppT (ConT ''ConstraintsFor) t) (VarT c)
+  let constraints = flip map ts $ \case
+        Left t -> AppT (AppT (ConT ''Has) (VarT c)) t
         Right t -> (AppT (VarT c) t)
-      l = length xs
-      constraints = foldl AppT (TupleT l) xs
-  [d| instance ArgDict $(varT c) $(pure typeHead) where
-        type ConstraintsFor  $(pure typeHead) $(varT c) = $(pure constraints)
-        argDict = $(LamCaseE <$> matches c constrs 'argDict)
-    |]
+  ms <- matches c constrs 'argDict
+  return
+    [ InstanceD Nothing constraints (AppT (AppT (ConT ''Has) (VarT c)) typeHead)
+      [ ValD (VarP 'argDict) (NormalB (LamCaseE ms)) [] ]
+    ]
 
 {-# DEPRECATED deriveArgDictV "Just use 'deriveArgDict'" #-}
 deriveArgDictV :: Name -> Q [Dec]
@@ -33,13 +32,13 @@ deriveArgDictV = deriveArgDict
 matches :: Name -> [Con] -> Name -> Q [Match]
 matches c constrs argDictName = do
   x <- newName "x"
-  fmap concat $ forM constrs $ \case
+  fmap concat . forM constrs $ \case
     GadtC [name] _ _ -> return $
       [Match (RecP name []) (NormalB $ ConE 'Dict) []]
     ForallC _ _ (GadtC [name] bts (AppT _ (VarT b))) -> do
       ps <- forM bts $ \case
         (_, AppT t (VarT b')) | b == b' -> do
-          hasArgDictInstance <- not . null <$> reifyInstances ''ArgDict [VarT c, t]
+          hasArgDictInstance <- not . null <$> reifyInstances ''Has [VarT c, t]
           return $ if hasArgDictInstance
             then Just x
             else Nothing
@@ -81,8 +80,11 @@ getDeclInfo n = reify n >>= \case
     tyVars <- replicateM (arity - 1) (newName "a")
     let typeHead = foldr (\v x -> AppT x (VarT v)) (ConT n) tyVars
     return (typeHead, constrs)
-  DataConI _ (ForallT _ _ (AppT typeHead _)) parent -> do
-    reify parent >>= \case
+  DataConI _ (AppT typeHead _) parent -> handleParent typeHead parent
+  DataConI _ (ForallT _ _ (AppT typeHead _)) parent -> handleParent typeHead parent
+  a -> error $ "getDeclInfo: Unmatched 'Info': " ++ show a
+  where
+    handleParent typeHead parent = reify parent >>= \case
       FamilyI _ instances -> do
         let instCons :: InstanceDec -> [Con]
             instCons = \case
@@ -103,14 +105,13 @@ getDeclInfo n = reify n >>= \case
           l@(_:_:_) -> error $ "getDeclInfo: Expected one data family instance for constructor " ++ show n ++ " but found multiple: " ++ show l
           [i] -> return (typeHead, instCons i)
       a -> error $ "getDeclInfo: Unmatched parent of data family instance: " ++ show a
-  a -> error $ "getDeclInfo: Unmatched 'Info': " ++ show a
 
 gadtIndices :: Name -> [Con] -> Q [Either Type Type]
 gadtIndices c constrs = fmap concat $ forM constrs $ \case
   GadtC _ _ (AppT _ typ) -> return [Right typ]
   ForallC _ _ (GadtC _ bts (AppT _ (VarT _))) -> fmap concat $ forM bts $ \case
     (_, AppT t (VarT _)) -> do
-      hasArgDictInstance <- fmap (not . null) $ reifyInstances ''ArgDict [VarT c, t]
+      hasArgDictInstance <- fmap (not . null) $ reifyInstances ''Has [VarT c, t]
       return $ if hasArgDictInstance then [Left t] else []
     _ -> return []
   ForallC _ _ (GadtC _ _ (AppT _ typ)) -> return [Right typ]
